@@ -49,6 +49,7 @@ void StartupClients(void)
    clientCount = 0;
    activeClient = NULL;
    currentDesktop = 0;
+   previousDesktop = 0;
 
    /* Clear out the client lists. */
    for(x = 0; x < LAYER_COUNT; x++) {
@@ -154,7 +155,7 @@ ClientNode *AddClientWindow(Window w, char alreadyMapped, char notOwner)
    np->state.defaultLayer = LAYER_NORMAL;
 
    np->state.border = BORDER_DEFAULT;
-   np->borderAction = BA_NONE;
+   np->mouseContext = MC_NONE;
 
    ReadClientInfo(np, alreadyMapped);
 
@@ -180,8 +181,6 @@ ClientNode *AddClientWindow(Window w, char alreadyMapped, char notOwner)
    }
    nodes[np->state.layer] = np;
 
-   SetDefaultCursor(np->window);
-
    if(notOwner) {
       XSetWindowAttributes sattr;
       JXAddToSaveSet(display, np->window);
@@ -1261,6 +1260,9 @@ void RemoveClient(ClientNode *np)
    if(np->className) {
       JXFree(np->className);
    }
+   if(np->machineName) {
+      Release(np->machineName);
+   }
 
    RemoveClientFromTaskBar(np);
    RemoveClientStrut(np);
@@ -1490,8 +1492,17 @@ void SignalUrgent(const TimeType *now, int x, int y, Window w, void *data)
 void UnmapClient(ClientNode *np)
 {
    if(np->state.status & STAT_MAPPED) {
+      XEvent e;
+
+      /* Unmap the window and record that we did so. */
       np->state.status &= ~STAT_MAPPED;
       JXUnmapWindow(display, np->window);
+
+      /* Discard the unmap event so we don't process it later. */
+      JXSync(display, False);
+      if(JXCheckTypedWindowEvent(display, np->window, UnmapNotify, &e)) {
+         UpdateTime(&e);
+      }
    }
 }
 
