@@ -19,7 +19,7 @@
 #include "swallow.h"
 #include "pager.h"
 #include "error.h"
-#include "key.h"
+#include "binding.h"
 #include "main.h"
 #include "font.h"
 #include "icon.h"
@@ -32,58 +32,75 @@
 #include "spacer.h"
 #include "desktop.h"
 #include "border.h"
+#include "default.h"
 
 #include <sys/types.h>
 #include <sys/stat.h>
-#include <sys/mman.h>
 #include <fcntl.h>
 #include <errno.h>
 
+/** Mapping of action names to values.
+ * Note that this mapping must be sorted.
+ */
+static const StringMappingType ACTION_MAP[] = {
+   { "center",                ACTION_CENTER        },
+   { "close",                 ACTION_CLOSE         },
+   { "ddesktop",              ACTION_DDESKTOP      },
+   { "desktop#",              ACTION_DESKTOP       },
+   { "down",                  ACTION_DOWN          },
+   { "escape",                ACTION_ESC           },
+   { "exit",                  ACTION_EXIT          },
+   { "fullscreen",            ACTION_FULLSCREEN    },
+   { "ldesktop",              ACTION_LDESKTOP      },
+   { "left",                  ACTION_LEFT          },
+   { "maxbottom",             ACTION_MAXBOTTOM     },
+   { "maxh",                  ACTION_MAXH          },
+   { "maximize",              ACTION_MAX           },
+   { "maxleft",               ACTION_MAXLEFT       },
+   { "maxright",              ACTION_MAXRIGHT      },
+   { "maxtop",                ACTION_MAXTOP        },
+   { "maxv",                  ACTION_MAXV          },
+   { "minimize",              ACTION_MIN           },
+   { "move",                  ACTION_MOVE          },
+   { "next",                  ACTION_NEXT          },
+   { "nextstacked",           ACTION_NEXTSTACK     },
+   { "none",                  ACTION_NONE          },
+   { "prev",                  ACTION_PREV          },
+   { "prevstacked",           ACTION_PREVSTACK     },
+   { "rdesktop",              ACTION_RDESKTOP      },
+   { "resize",                ACTION_RESIZE        },
+   { "restart",               ACTION_RESTART       },
+   { "restore",               ACTION_RESTORE       },
+   { "right",                 ACTION_RIGHT         },
+   { "select",                ACTION_ENTER         },
+   { "send#",                 ACTION_SEND          },
+   { "sendd",                 ACTION_SENDD         },
+   { "sendl",                 ACTION_SENDL         },
+   { "sendr",                 ACTION_SENDR         },
+   { "sendu",                 ACTION_SENDU         },
+   { "shade",                 ACTION_SHADE         },
+   { "showdesktop",           ACTION_SHOWDESK      },
+   { "showtray",              ACTION_SHOWTRAY      },
+   { "stick",                 ACTION_STICK         },
+   { "udesktop",              ACTION_UDESKTOP      },
+   { "up",                    ACTION_UP            },
+   { "window",                ACTION_WIN           }
+};
+static const unsigned int ACTION_MAP_COUNT = ARRAY_LENGTH(ACTION_MAP);
+
 /** Mapping of key names to key types.
  * Note that this mapping must be sorted.
  */
-static const StringMappingType KEY_MAP[] = {
-   { "close",                 KEY_CLOSE         },
-   { "ddesktop",              KEY_DDESKTOP      },
-   { "desktop#",              KEY_DESKTOP       },
-   { "down",                  KEY_DOWN          },
-   { "escape",                KEY_ESC           },
-   { "exit",                  KEY_EXIT          },
-   { "fullscreen",            KEY_FULLSCREEN    },
-   { "ldesktop",              KEY_LDESKTOP      },
-   { "left",                  KEY_LEFT          },
-   { "maxbottom",             KEY_MAXBOTTOM     },
-   { "maxh",                  KEY_MAXH          },
-   { "maximize",              KEY_MAX           },
-   { "maxleft",               KEY_MAXLEFT       },
-   { "maxright",              KEY_MAXRIGHT      },
-   { "maxtop",                KEY_MAXTOP        },
-   { "maxv",                  KEY_MAXV          },
-   { "minimize",              KEY_MIN           },
-   { "move",                  KEY_MOVE          },
-   { "next",                  KEY_NEXT          },
-   { "nextstacked",           KEY_NEXTSTACK     },
-   { "prev",                  KEY_PREV          },
-   { "prevstacked",           KEY_PREVSTACK     },
-   { "rdesktop",              KEY_RDESKTOP      },
-   { "resize",                KEY_RESIZE        },
-   { "restart",               KEY_RESTART       },
-   { "restore",               KEY_RESTORE       },
-   { "right",                 KEY_RIGHT         },
-   { "select",                KEY_ENTER         },
-   { "sendd",                 KEY_SENDD         },
-   { "sendl",                 KEY_SENDL         },
-   { "sendr",                 KEY_SENDR         },
-   { "sendu",                 KEY_SENDU         },
-   { "shade",                 KEY_SHADE         },
-   { "showdesktop",           KEY_SHOWDESK      },
-   { "showtray",              KEY_SHOWTRAY      },
-   { "stick",                 KEY_STICK         },
-   { "udesktop",              KEY_UDESKTOP      },
-   { "up",                    KEY_UP            },
-   { "window",                KEY_WIN           }
+static const StringMappingType MC_MAP[] = {
+   { "border",                MC_BORDER   },
+   { "close",                 MC_CLOSE    },
+   { "icon",                  MC_ICON     },
+   { "maximize",              MC_MAXIMIZE },
+   { "minimize",              MC_MINIMIZE },
+   { "root",                  MC_ROOT     },
+   { "title",                 MC_MOVE     }
 };
-static const unsigned int KEY_MAP_COUNT = ARRAY_LENGTH(KEY_MAP);
+static const unsigned int MC_MAP_COUNT = ARRAY_LENGTH(MC_MAP);
 
 /** Mapping of names to group options.
  * Note that this mapping must be sorted.
@@ -109,6 +126,8 @@ static const StringMappingType OPTION_MAP[] = {
    { "nofullscreen",       OPTION_NOFULLSCREEN  },
    { "nolist",             OPTION_NOLIST        },
    { "nomax",              OPTION_NOMAX         },
+   { "nomaxborder",        OPTION_NOMAXBORDER   },
+   { "nomaxtitle",         OPTION_NOMAXTITLE    },
    { "nomin",              OPTION_NOMIN         },
    { "nomove",             OPTION_NOMOVE        },
    { "nopager",            OPTION_NOPAGER       },
@@ -117,6 +136,7 @@ static const StringMappingType OPTION_MAP[] = {
    { "notitle",            OPTION_NOTITLE       },
    { "noturgent",          OPTION_NOTURGENT     },
    { "pignore",            OPTION_PIGNORE       },
+   { "shaded",             OPTION_SHADED        },
    { "sticky",             OPTION_STICKY        },
    { "tiled",              OPTION_TILED         },
    { "title",              OPTION_TITLE         },
@@ -135,16 +155,27 @@ static const char *X_ATTRIBUTE = "x";
 static const char *Y_ATTRIBUTE = "y";
 static const char *WIDTH_ATTRIBUTE = "width";
 static const char *HEIGHT_ATTRIBUTE = "height";
+static const char *SCREEN_ATTRIBUTE = "screen";
 static const char *DYNAMIC_ATTRIBUTE = "dynamic";
 static const char *SPACING_ATTRIBUTE = "spacing";
+static const char *TIMEOUT_ATTRIBUTE = "timeout";
+static const char *POPUP_ATTRIBUTE = "popup";
 
 static const char *FALSE_VALUE = "false";
 static const char *TRUE_VALUE = "true";
 
+static const char * const CONFIG_FILES[] = {
+  "$XDG_CONFIG_HOME/jwm/jwmrc",
+  "$HOME/.config/jwm/jwmrc",
+  "$HOME/.jwmrc",
+  SYSTEM_CONFIG
+};
+static const unsigned CONFIG_FILE_COUNT = ARRAY_LENGTH(CONFIG_FILES);
+
+static void ParseInternal(const char *config);
 static char ParseFile(const char *fileName, int depth);
-static char *ReadFile(FILE *fd);
 static TokenNode *TokenizeFile(const char *fileName);
-static TokenNode *TokenizePipe(const char *command);
+static TokenNode *TokenizePipe(const char *command, unsigned timeout_ms);
 
 /* Misc. */
 static void Parse(const TokenNode *start, int depth);
@@ -162,6 +193,7 @@ static MenuItem *InsertMenuItem(MenuItem *last);
 static MenuItem *ParseMenuInclude(const TokenNode *tp, Menu *menu,
                                   MenuItem *last);
 static TokenNode *ParseMenuIncludeHelper(const TokenNode *tp,
+                                         unsigned timeout_ms,
                                          const char *command);
 
 /* Tray. */
@@ -198,7 +230,9 @@ static void ParseMenuStyle(const TokenNode *tp);
 static void ParsePopupStyle(const TokenNode *tp);
 
 /* Feel. */
+static ActionType ParseAction(const char *str, const char **command);
 static void ParseKey(const TokenNode *tp);
+static void ParseMouse(const TokenNode *tp);
 static void ParseSnapMode(const TokenNode *tp);
 static void ParseMoveMode(const TokenNode *tp);
 static void ParseResizeMode(const TokenNode *tp);
@@ -213,22 +247,44 @@ static int ParseTokenValue(const StringMappingType *mapping, int count,
 static int ParseAttribute(const StringMappingType *mapping, int count,
                           const TokenNode *tp, const char *attr,
                           int def);
-static unsigned int ParseUnsigned(const TokenNode *tp, const char *str);
+static int ParseSigned(const TokenNode *tp, const char *str);
+static unsigned ParseUnsigned(const TokenNode *tp, const char *str);
+static unsigned ParseTimeout(const TokenNode *tp, unsigned timeout_ms);
 static unsigned int ParseOpacity(const TokenNode *tp, const char *str);
 static WinLayerType ParseLayer(const TokenNode *tp, const char *str);
 static StatusWindowType ParseStatusWindowType(const TokenNode *tp);
 static void InvalidTag(const TokenNode *tp, TokenType parent);
 static void ParseError(const TokenNode *tp, const char *str, ...);
 
+/** Parse the default configuration. */
+void ParseInternal(const char *config)
+{
+   TokenNode *tokens = Tokenize(config, "");
+   Parse(tokens, 0);
+   ReleaseTokens(tokens);
+}
+
 /** Parse the JWM configuration. */
 void ParseConfig(const char *fileName)
 {
-   if(!ParseFile(fileName, 0)) {
-      if(JUNLIKELY(!ParseFile(SYSTEM_CONFIG, 0))) {
-         ParseError(NULL, _("could not open %s or %s"),
-                    fileName, SYSTEM_CONFIG);
+   ParseInternal(BASE_CONFIG);
+   if(fileName) {
+      if(!ParseFile(fileName, 0)) {
+         ParseError(NULL, _("could not open %s"), fileName);
+         ParseInternal(DEFAULT_CONFIG);
       }
+   } else {
+      unsigned i;
+      for(i = 0; i < CONFIG_FILE_COUNT; i++) {
+         if(ParseFile(CONFIG_FILES[i], 0)) {
+            goto ConfigFileFound;
+         }
+      }
+      ParseError(NULL, _("could not open %s or %s"),
+                 CONFIG_FILES[0], SYSTEM_CONFIG);
+      ParseInternal(DEFAULT_CONFIG);
    }
+ConfigFileFound:
    ValidateTrayButtons();
    ValidateKeys();
 }
@@ -308,6 +364,9 @@ void Parse(const TokenNode *start, int depth)
             case TOK_KEY:
                ParseKey(tp);
                break;
+            case TOK_MOUSE:
+               ParseMouse(tp);
+               break;
             case TOK_MENUSTYLE:
                ParseMenuStyle(tp);
                break;
@@ -374,6 +433,9 @@ void Parse(const TokenNode *start, int depth)
             case TOK_DEFAULTICON:
                SetDefaultIcon(tp->value);
                break;
+            case TOK_TITLEBUTTONORDER:
+               SetTitleButtonOrder(tp->value);
+               break;
             default:
                InvalidTag(tp, TOK_JWM);
                break;
@@ -500,9 +562,9 @@ void ParseRootMenu(const TokenNode *start)
 
    value = FindAttribute(start->attributes, DYNAMIC_ATTRIBUTE);
    menu->dynamic = CopyString(value);
+   menu->timeout_ms = ParseTimeout(start, MENU_TIMEOUT_MS);
 
    SetRootMenu(onroot, menu);
-
 }
 
 /** Insert a new menu item into a menu. */
@@ -545,6 +607,7 @@ MenuItem *ParseMenuItem(const TokenNode *start, Menu *menu, MenuItem *last)
 
          last->action.type = MA_DYNAMIC;
          last->action.str = CopyString(start->value);
+         last->action.timeout_ms = ParseTimeout(start, MENU_TIMEOUT_MS);
 
          value = FindAttribute(start->attributes, HEIGHT_ATTRIBUTE);
          if(value) {
@@ -759,12 +822,13 @@ MenuItem *ParseMenuItem(const TokenNode *start, Menu *menu, MenuItem *last)
 }
 
 /** Get tokens from a menu include (either dynamic or static). */
-TokenNode *ParseMenuIncludeHelper(const TokenNode *tp, const char *command)
+TokenNode *ParseMenuIncludeHelper(const TokenNode *tp,
+                                  unsigned timeout_ms,
+                                  const char *command)
 {
    TokenNode *start;
-
    if(!strncmp(command, "exec:", 5)) {
-      start = TokenizePipe(&command[5]);
+      start = TokenizePipe(&command[5], timeout_ms);
    } else {
       start = TokenizeFile(command);
    }
@@ -783,7 +847,10 @@ TokenNode *ParseMenuIncludeHelper(const TokenNode *tp, const char *command)
 MenuItem *ParseMenuInclude(const TokenNode *tp, Menu *menu,
                            MenuItem *last)
 {
-   TokenNode *start = ParseMenuIncludeHelper(tp, tp->value);
+   TokenNode *start;
+   const unsigned timeout_ms = ParseTimeout(tp, INCLUDE_TIMEOUT_MS);
+
+   start = ParseMenuIncludeHelper(tp, timeout_ms, tp->value);
    if(JLIKELY(start)) {
       last = ParseMenuItem(start->subnodeHead, menu, last);
       ReleaseTokens(start);
@@ -792,10 +859,10 @@ MenuItem *ParseMenuInclude(const TokenNode *tp, Menu *menu,
 }
 
 /** Parse a dynamic menu (called from menu code). */
-Menu *ParseDynamicMenu(const char *command)
+Menu *ParseDynamicMenu(unsigned timeout_ms, const char *command)
 {
    Menu *menu = NULL;
-   TokenNode *start = ParseMenuIncludeHelper(NULL, command);
+   TokenNode *start = ParseMenuIncludeHelper(NULL, timeout_ms, command);
    if(JLIKELY(start)) {
       menu = ParseMenu(start);
       ReleaseTokens(start);
@@ -803,17 +870,50 @@ Menu *ParseDynamicMenu(const char *command)
    return menu;
 }
 
-/** Parse a key binding. */
-void ParseKey(const TokenNode *tp) {
+/** Parse an action. */
+ActionType ParseAction(const char *str, const char **command)
+{
+   ActionType action;
+   action.action = ACTION_INVALID;
+   action.extra = 0;
+   *command = NULL;
+   if(!strncmp(str, "exec:", 5)) {
+      action.action = ACTION_EXEC;
+      *command = str + 5;
+   } else if(!strncmp(str, "root:", 5)) {
+      action.action = ACTION_ROOT;
+      *command = str + 5;
+   } else if(!strncmp(str, "resize:", 7)) {
+      int i;
+      action.action = ACTION_RESIZE;
+      for(i = 7; str[i]; i++) {
+         switch(str[i]) {
+         case 'n': action.extra |= ACTION_RESIZE_N; break;
+         case 's': action.extra |= ACTION_RESIZE_S; break;
+         case 'e': action.extra |= ACTION_RESIZE_E; break;
+         case 'w': action.extra |= ACTION_RESIZE_W; break;
+         default: break;
+         }
+      }
+   } else {
+      /* Look up the option in the key map using binary search. */
+      const int x = FindValue(ACTION_MAP, ACTION_MAP_COUNT, str);
+      if(x >= 0) {
+         action.action = (unsigned char)x;
+      }
+   }
+   return action;
+}
 
+/** Parse a key binding. */
+void ParseKey(const TokenNode *tp)
+{
    const char *key;
    const char *code;
    const char *mask;
    const char *action;
    const char *command;
-   KeyType k;
-
-   Assert(tp);
+   ActionType k;
 
    mask = FindAttribute(tp->attributes, "mask");
    key = FindAttribute(tp->attributes, "key");
@@ -825,41 +925,59 @@ void ParseKey(const TokenNode *tp) {
       return;
    }
 
-   command = NULL;
-   k = KEY_NONE;
-   if(!strncmp(action, "exec:", 5)) {
-      k = KEY_EXEC;
-      command = action + 5;
-   } else if(!strncmp(action, "root:", 5)) {
-      k = KEY_ROOT;
-      command = action + 5;
-   } else {
-      /* Look up the option in the key map using binary search. */
-      const int x = FindValue(KEY_MAP, KEY_MAP_COUNT, action);
-      if(x >= 0) {
-         k = (KeyType)x;
-      }
-   }
-
    /* Insert the binding if it's valid. */
-   if(JUNLIKELY(k == KEY_NONE)) {
-
+   k = ParseAction(action, &command);
+   if(JUNLIKELY(k.action == ACTION_INVALID)) {
       ParseError(tp, _("invalid Key action: \"%s\""), action);
-
    } else {
-
       InsertBinding(k, mask, key, code, command);
+   }
+}
 
+/** Parse a mouse binding. */
+void ParseMouse(const TokenNode *tp)
+{
+   const char *mask;
+   const char *context;
+   const char *action;
+   const char *command;
+   int button;
+   int x;
+   ActionType key;
+
+   button = ParseSigned(tp, FindAttribute(tp->attributes, "button"));
+   if(JUNLIKELY(button == 0)) {
+      return;
+   }
+   mask = FindAttribute(tp->attributes, "mask");
+   context = FindAttribute(tp->attributes, "context");
+
+   action = tp->value;
+   if(JUNLIKELY(action == NULL)) {
+      ParseError(tp, _("no action specified for Mouse"));
+      return;
+   }
+   key = ParseAction(action, &command);
+   if(JUNLIKELY(key.action == ACTION_INVALID)) {
+      ParseError(tp, _("invalid Mouse action: \"%s\""), action);
+      return;
    }
 
+   x = FindValue(MC_MAP, MC_MAP_COUNT, context);
+   if(JUNLIKELY(x < 0)) {
+      ParseError(tp, _("invalid Mouse context: \"%s\""), context);
+   } else {
+      const MouseContextType mc = (MouseContextType)x;
+      InsertMouseBinding(button, mask, mc, key, command);
+   }
 }
 
 /** Parse text alignment. */
 AlignmentType ParseTextAlignment(const TokenNode *tp)
 {
    static const StringMappingType mapping[] = {
-      {"left",    ALIGN_LEFT   },
       {"center",  ALIGN_CENTER },
+      {"left",    ALIGN_LEFT   },
       {"right",   ALIGN_RIGHT  }
    };
    const char *attr= FindAttribute(tp->attributes, "align");
@@ -953,13 +1071,16 @@ void ParseActiveWindowStyle(const TokenNode *tp)
 /** Parse an include. */
 void ParseInclude(const TokenNode *tp, int depth)
 {
+   unsigned timeout_ms;
+
    if(JUNLIKELY(!tp->value)) {
       ParseError(tp, _("no include file specified"));
       return;
    }
 
+   timeout_ms = ParseTimeout(tp, INCLUDE_TIMEOUT_MS);
    if(!strncmp(tp->value, "exec:", 5)) {
-      TokenNode *tokens = TokenizePipe(&tp->value[5]);
+      TokenNode *tokens = TokenizePipe(&tp->value[5], timeout_ms);
       if(JLIKELY(tokens)) {
          Parse(tokens, 0);
          ReleaseTokens(tokens);
@@ -976,14 +1097,23 @@ void ParseInclude(const TokenNode *tp, int depth)
 
 /** Parse desktop configuration. */
 void ParseDesktops(const TokenNode *tp) {
-
+   
+   static const StringMappingType mapping[] = {
+      { "off",       DBACKANDFORTH_OFF },
+      { "on",        DBACKANDFORTH_ON  }
+   };
    TokenNode *np;
    const char *width;
    const char *height;
    int desktop;
+   DesktopBackAndForthType backandforth;
 
    Assert(tp);
 
+   backandforth = ParseAttribute(mapping, ARRAY_LENGTH(mapping), tp,
+                                 "backandforth", DBACKANDFORTH_OFF);
+   settings.desktopBackAndForth = backandforth;
+
    width = FindAttribute(tp->attributes, WIDTH_ATTRIBUTE);
    if(width != NULL) {
       settings.desktopWidth = ParseUnsigned(tp, width);
@@ -1060,10 +1190,10 @@ void ParseTrayStyle(const TokenNode *tp, FontType font, ColorType fg)
    const ColorType activeUp = fg + COLOR_TRAY_ACTIVE_UP - COLOR_TRAY_FG;
    const ColorType activeDown = fg + COLOR_TRAY_ACTIVE_DOWN - COLOR_TRAY_FG;
 
-   /* TrayStyle has extra attributes. */
-   if(tp->type == TOK_TRAYSTYLE) {
+   /* TaskListStyle has extra attributes. */
+   if(tp->type == TOK_TASKLISTSTYLE) {
       const char *temp;
-      ParseDecorations(tp, &settings.trayDecorations);
+      ParseDecorations(tp, &settings.taskListDecorations);
       temp = FindAttribute(tp->attributes, "group");
       if(temp) {
          settings.groupTasks = !strcmp(temp, TRUE_VALUE);
@@ -1073,8 +1203,8 @@ void ParseTrayStyle(const TokenNode *tp, FontType font, ColorType fg)
       if(temp) {
          settings.listAllTasks = !strcmp(temp, "all");
       }
-   } else if(tp->type == TOK_TASKLISTSTYLE) {
-      ParseDecorations(tp, &settings.taskListDecorations);
+   } else if(tp->type == TOK_TRAYSTYLE) {
+      ParseDecorations(tp, &settings.trayDecorations);
    }
 
    for(np = tp->subnodeHead; np; np = np->next) {
@@ -1144,14 +1274,17 @@ void ParseActive(const TokenNode *tp, ColorType fg,
 void ParseTray(const TokenNode *tp)
 {
    static const StringMappingType mapping[] = {
-      { "bottom",    THIDE_BOTTOM   },
-      { "left",      THIDE_LEFT     },
-      { "off",       THIDE_OFF      },
-      { "right",     THIDE_RIGHT    },
-      { "top",       THIDE_TOP      }
+      { "bottom",    THIDE_BOTTOM      },
+      { "invisible", THIDE_INVISIBLE   },
+      { "left",      THIDE_LEFT        },
+      { "off",       THIDE_OFF         },
+      { "on",        THIDE_ON          },
+      { "right",     THIDE_RIGHT       },
+      { "top",       THIDE_TOP         }
    };
    const TokenNode *np;
    const char *attr;
+   unsigned delay = 0;
    TrayType *tray;
    TrayAutoHideType autohide;
 
@@ -1161,7 +1294,11 @@ void ParseTray(const TokenNode *tp)
 
    autohide = ParseAttribute(mapping, ARRAY_LENGTH(mapping), tp,
                              "autohide", THIDE_OFF);
-   SetAutoHideTray(tray, autohide);
+   attr = FindAttribute(tp->attributes, "delay");
+   if(attr) {
+      delay = ParseUnsigned(tp, attr);
+   }
+   SetAutoHideTray(tray, autohide, delay);
 
    attr = FindAttribute(tp->attributes, X_ATTRIBUTE);
    if(attr) {
@@ -1183,6 +1320,11 @@ void ParseTray(const TokenNode *tp)
       SetTrayHeight(tray, attr);
    }
 
+   attr = FindAttribute(tp->attributes, SCREEN_ATTRIBUTE);
+   if(attr) {
+      SetTrayScreen(tray, attr);
+   }
+
    attr = FindAttribute(tp->attributes, "valign");
    SetTrayVerticalAlignment(tray, attr);
 
@@ -1361,7 +1503,7 @@ void ParseTrayButton(const TokenNode *tp, TrayType *tray)
 
    icon = FindAttribute(tp->attributes, ICON_ATTRIBUTE);
    label = FindAttribute(tp->attributes, LABEL_ATTRIBUTE);
-   popup = FindAttribute(tp->attributes, "popup");
+   popup = FindAttribute(tp->attributes, POPUP_ATTRIBUTE);
 
    temp = FindAttribute(tp->attributes, WIDTH_ATTRIBUTE);
    if(temp) {
@@ -1696,6 +1838,12 @@ void ParseGroup(const TokenNode *tp)
       case TOK_NAME:
          AddGroupName(group, np->value);
          break;
+      case TOK_TYPE:
+         AddGroupType(group, np->value);
+         break;
+      case TOK_MACHINE:
+         AddGroupMachine(group, np->value);
+         break;
       case TOK_OPTION:
          ParseGroupOption(np, group, np->value);
          break;
@@ -1857,48 +2005,6 @@ int ParseAttribute(const StringMappingType *mapping, int count,
    }
 }
 
-/** Read a file. */
-char *ReadFile(FILE *fd)
-{
-   const int BLOCK_SIZE = 1024;
-
-   char *buffer;
-   int len, max;
-
-   len = 0;
-   max = BLOCK_SIZE;
-   buffer = Allocate(max + 1);
-
-   for(;;) {
-      const size_t count = fread(&buffer[len], 1, max - len, fd);
-      if(count == 0) {
-         if(feof(fd)) {
-            break;
-         } else if(errno == EAGAIN || errno == EWOULDBLOCK || errno == EINTR) {
-            continue;
-         } else {
-            Warning(_("could not read file: %s"), strerror(errno));
-            break;
-         }
-      }
-      len += count;
-      if(len == max) {
-         max += BLOCK_SIZE;
-         if(JUNLIKELY(max < 0)) {
-            /* File is too big. */
-            break;
-         }
-         buffer = Reallocate(buffer, max + 1);
-         if(JUNLIKELY(buffer == NULL)) {
-            FatalError(_("out of memory"));
-         }
-      }
-   }
-   buffer[len] = 0;
-
-   return buffer;
-}
-
 /** Tokenize a file by memory mapping it. */
 TokenNode *TokenizeFile(const char *fileName)
 {
@@ -1906,6 +2012,7 @@ TokenNode *TokenizeFile(const char *fileName)
    TokenNode *tokens;
    char *path;
    char *buffer;
+   ssize_t offset;
 
    path = CopyString(fileName);
    ExpandPath(&path);
@@ -1920,36 +2027,34 @@ TokenNode *TokenizeFile(const char *fileName)
       close(fd);
       return NULL;
    }
-   buffer = mmap(NULL, sbuf.st_size, PROT_READ, MAP_SHARED, fd, 0);
-   if(JUNLIKELY(buffer == MAP_FAILED)) {
-      close(fd);
-      return NULL;
+   buffer = Allocate(sbuf.st_size + 1);
+   offset = 0;
+   while(offset < sbuf.st_size) {
+      const ssize_t rc = read(fd, &buffer[offset], sbuf.st_size - offset);
+      if(rc <= 0) {
+         break;
+      }
+      offset += rc;
    }
+   buffer[offset] = 0;
    tokens = Tokenize(buffer, fileName);
-   munmap(buffer, sbuf.st_size);
+   Release(buffer);
    close(fd);
    return tokens;
 }
 
 /** Tokenize the output of a command. */
-TokenNode *TokenizePipe(const char *command)
+TokenNode *TokenizePipe(const char *command, unsigned timeout_ms)
 {
    TokenNode *tokens;
-   FILE *fp;
    char *path;
    char *buffer;
 
    path = CopyString(command);
    ExpandPath(&path);
 
-   fp = popen(path, "r");
+   buffer = ReadFromProcess(path, timeout_ms);
    Release(path);
-
-   buffer = NULL;
-   if(JLIKELY(fp)) {
-      buffer = ReadFile(fp);
-      pclose(fp);
-   }
    if(JUNLIKELY(!buffer)) {
       return NULL;
    }
@@ -1959,8 +2064,25 @@ TokenNode *TokenizePipe(const char *command)
    return tokens;
 }
 
+/** Parse a signed integer. */
+int ParseSigned(const TokenNode *tp, const char *str)
+{
+   long value;
+   if(JUNLIKELY(!str)) {
+      ParseError(tp, _("no value specified"));
+      return 0;
+   }
+   value = strtol(str, NULL, 0);
+   if(JUNLIKELY(value < INT_MIN || value > INT_MAX)) {
+      ParseError(tp, _("invalid setting: %s"), str);
+      return 0;
+   } else {
+      return (int)value;
+   }
+}
+
 /** Parse an unsigned integer. */
-unsigned int ParseUnsigned(const TokenNode *tp, const char *str)
+unsigned ParseUnsigned(const TokenNode *tp, const char *str)
 {
    long value;
    if(JUNLIKELY(!str)) {
@@ -1972,14 +2094,25 @@ unsigned int ParseUnsigned(const TokenNode *tp, const char *str)
       ParseError(tp, _("invalid setting: %s"), str);
       return 0;
    } else {
-      return (unsigned int)value;
+      return (unsigned)value;
    }
 }
 
-/** Parse opacity (a float between 0.0 and 1.0). */
-unsigned int ParseOpacity(const TokenNode *tp, const char *str)
+/** Parse a timeout attribute. */
+unsigned ParseTimeout(const TokenNode *tp, unsigned timeout_ms)
 {
-   float value;
+   char *temp = FindAttribute(tp->attributes, TIMEOUT_ATTRIBUTE);
+   if(temp) {
+      const unsigned ms = ParseUnsigned(tp, temp);
+      timeout_ms = ms == 0 ? timeout_ms : ms;
+   }
+   return timeout_ms;
+}
+
+/** Parse opacity (a float between 0.0 and 1.0). */
+unsigned ParseOpacity(const TokenNode *tp, const char *str)
+{
+   double value;
    if(JUNLIKELY(!str)) {
       ParseError(tp, _("no value specified"));
       return UINT_MAX;
