@@ -62,6 +62,32 @@ typedef unsigned char AlignmentType;
 #define ALIGN_CENTER    1
 #define ALIGN_RIGHT     2
 
+/** Mouse binding contexts. */
+typedef unsigned char MouseContextType;
+#define MC_NONE            0     /**< Keyboard/none. */
+#define MC_ROOT            1     /**< Root window. */
+#define MC_BORDER          2     /**< Resize handle. */
+#define MC_MOVE            3     /**< Move handle. */
+#define MC_CLOSE           4     /**< Close button. */
+#define MC_MAXIMIZE        5     /**< Maximize button. */
+#define MC_MINIMIZE        6     /**< Minimize button. */
+#define MC_ICON            7     /**< Window menu button. */
+#define MC_COUNT           8     /**< Number of contexts. */
+#define MC_MASK            0x0F  /**< Context type mask. */
+#define MC_BORDER_N        0x10  /**< North border. */
+#define MC_BORDER_S        0x20  /**< South border. */
+#define MC_BORDER_E        0x40  /**< East border. */
+#define MC_BORDER_W        0x80  /**< West border. */
+
+/** Enumeration of desktop back and forth values. */
+typedef unsigned char DesktopBackAndForthType;
+#define DBACKANDFORTH_OFF 0 /**< No back and forth */
+#define DBACKANDFORTH_ON  1 /**< Enable back and forth */
+
+/** Maximimum number of title bar components
+ * For now, we allow each component to be used twice. */
+#define TBC_COUNT       9
+
 /** Settings. */
 typedef struct {
    unsigned doubleClickSpeed;
@@ -93,8 +119,10 @@ typedef struct {
    DecorationsType taskListDecorations;
    DecorationsType menuDecorations;
    PopupMaskType popupMask;
+   MouseContextType titleBarLayout[TBC_COUNT + 1];
    char groupTasks;
    char listAllTasks;
+   DesktopBackAndForthType desktopBackAndForth;
 } Settings;
 
 extern Settings settings;
@@ -109,4 +137,7 @@ void StartupSettings(void);
 /** Update a string setting. */
 void SetPathString(char **dest, const char *src);
 
+/** Set the title button order. */
+void SetTitleButtonOrder(const char *order);
+
 #endif
